# -*- coding: utf-8 -*-

import unittest
import tempfile
import os

import sqlist


class TestSQList(unittest.TestCase):
    def setUp(self):
        self.test_values = [
            {'foo': 'bar', 'baz': None},
            [1, 1, 2, 3, 5, 8, 13, 21, 34, 55],
            set(u'Eyjafjallajökull'),
            u'埃亚菲亚德拉冰盖'
        ]
        self.comparable_values = [
            'aaa',
            'ccc',
            'eee',
            'ddd',
            'bbb'
        ]
        self.sl = sqlist.SQList(self.test_values)

    def tearDown(self):
        del self.sl
        del self.test_values

    def test_constructor_creates_correct_table(self):
        """
        Constructor should create one table with two columns
        """
        sl = sqlist.SQList()
        expected_pragma = [
            {
                'name': 'key',
                'type': 'BLOB',
                'notnull': False
            },
            {
                'name': 'value',
                'type': 'BLOB',
                'notnull': True
            }
        ]
        pragma = [{'name': i[1], 'type': i[2], 'notnull': i[3]}
                  for i in sl.cursor.execute('''PRAGMA table_info(`data`);''')]
        self.assertEqual(len(pragma), len(expected_pragma))
        for table in expected_pragma:
            self.assertTrue(table in pragma)

    def test_constructor_creates_correct_indexes(self):
        """
        There should be only one not unique index on `key` column
        """
        sl = sqlist.SQList()
        index_list = list(sl.cursor.execute('''PRAGMA index_list(`data`);'''))
        self.assertEqual(len(index_list), 1)
        index_name = index_list[0][1]
        index_info = sl.cursor.execute(
            '''PRAGMA index_info(%s);''' % index_name).fetchone()
        self.assertEqual(index_info[2], 'key')

    def test_constructor_creates_file_of_database(self):
        temp_dir = tempfile.mkdtemp()
        temp_file = os.path.join(temp_dir, 'test_sqlist.db')
        sl = sqlist.SQList(path=temp_file)

        self.assertTrue(os.path.isfile(temp_file))

        os.remove(temp_file)
        os.removedirs(temp_dir)

    def test_constructor_with_list_of_predefined_values_and_contains_method(self):
        self.assertEqual(len(self.test_values), len(self.sl))
        for i in self.test_values:
            self.assertTrue(i in self.sl)
        self.assertFalse('some_spam_string' in self.sl)

    def test_constructor_with_key_parameter(self):
        def key(x):
            return len(x)

        self.assertEqual(list(sqlist.SQList(self.test_values, key=key)),
                         sorted(self.test_values, key=key))

    def test_eq_method(self):
        self.assertTrue(self.sl.__eq__(self.test_values))

    def test_iter_method(self):
        self.assertEqual(self.test_values, list(self.sl))

    def test_len_method(self):
        self.assertEqual(len(self.test_values), len(self.sl))

    def test_setitem_and_getitem_methods(self):
        for i in range(len(self.sl)):
            self.sl[i] = self.test_values[i]
            self.assertEqual(self.sl[i], self.test_values[i])

        self.assertEqual(self.sl[0:3], self.test_values[0:3])

        self.assertRaises(IndexError, self.sl.__getitem__, len(self.sl) + 5)
        self.assertRaises(IndexError, self.sl.__setitem__, len(self.sl) + 5, 3)

    def test_delitem_method(self):
        del self.sl[0]

        self.assertEqual(len(self.sl), len(self.test_values) - 1)
        self.assertFalse(self.test_values[0] in self.sl)

        self.assertRaises(IndexError, self.sl.__delitem__, len(self.sl) + 5)

    def test_append_method(self):
        test_appendix = (1, 2, 3)
        self.sl.append(test_appendix)

        self.assertEqual(len(self.sl), len(self.test_values) + 1)
        self.assertEqual(self.sl[-1], test_appendix)
        self.assertTrue(test_appendix in self.sl)

    def test_pop_method(self):
        self.assertEqual(self.sl.pop(), self.test_values.pop())
        self.assertEqual(self.sl.pop(0), self.test_values.pop(0))
        self.assertEqual(self.sl.pop(1), self.test_values.pop(1))
        self.assertEqual(len(self.test_values), len(self.sl))

    def test_sort_method_direct_with_key(self):
        def key(x):
            return len(x)
        sl = sqlist.SQList(self.comparable_values)

        sl.sort(key=key)
        self.comparable_values.sort(key=key)
        self.assertEqual(sl, self.comparable_values)
        self.assertTrue(sl.key is None)

    def test_sort_method_reversed_with_key(self):
        def key(x):
            return len(x)
        sl = sqlist.SQList(self.comparable_values)

        sl.sort(key=key, reverse=True)
        self.comparable_values.sort(key=key, reverse=True)
        self.assertEqual(sl, self.comparable_values)
        self.assertTrue(sl.key is None)

    def test_sort_method_direct_without_key(self):
        sl = sqlist.SQList(self.comparable_values, key=lambda x: hash(x))

        sl.sort()
        self.comparable_values.sort()
        self.assertEqual(list(sl), self.comparable_values)
        self.assertTrue(sl.key is None)

    def test_sort_method_reversed_without_key(self):
        sl = sqlist.SQList(self.comparable_values, key=lambda x: hash(x))

        sl.sort(reverse=True)
        self.comparable_values.sort(reverse=True)
        self.assertEqual(sl, self.comparable_values)
        self.assertTrue(sl.key is None)

    def test_extend_method(self):
        self.sl.extend(self.test_values)

        self.assertEqual(self.sl, self.test_values + self.test_values)

    def test_extend_method_with_noniterable(self):
        with self.assertRaises(TypeError):
            self.sl.extend(24)

    def test_temp_class(self):
        sl = sqlist.SQList.temp()
        self.assertTrue(os.path.exists(sl.path))



from setuptools import setup, find_packages

setup(
    name='sqlist',
    version='0.2.1',
    packages=find_packages(),
    description='List-like wrapper for SQLite',
    long_description=open('README.rst').read(),

    author='Valery Ryaboshapko',
    author_email='valera@creator.su',
    license='MIT',
    url='https://github.com/valericus/sqlist',

    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],

    zip_safe=True
)



# -*- coding: utf-8 -*-

import pickle
import sqlite3
import tempfile

from os import close

from collections import Iterable


class SQList(object):
    """
    List-like object that stores data in SQLite database
    """
    def __init__(self, values=None, path=':memory:', key=None, drop=True):
        """
        :param values: iterable with elements of new list
        :param path: path to file with SQLite database
        :param key: callable object used to sort values of SQList
        :param drop: drop any data in file at path
        """
        if key:
            if callable(key):
                self.key = key
            else:
                raise TypeError('%s object is not callable' % type(key))
        else:
            self.key = lambda x: None

        self.path = path
        self.sql = sqlite3.connect(path)
        self.sql.text_factory = str
        self.cursor = self.sql.cursor()
        if drop:
            self.cursor.execute('''DROP TABLE IF EXISTS `data`;''')
        self.cursor.execute(
            '''CREATE TABLE `data` (`key` BLOB,
                                    `value` BLOB NOT NULL);''')
        self.cursor.execute('''CREATE INDEX `keys_index` ON `data` (`key`);''')
        if values:
            self.cursor.executemany(
                '''INSERT INTO `data`
                   (`key`, `value`)
                   VALUES (?, ?);''',
                zip(map(self.key, values), map(pickle.dumps, values))
            )
            self.sql.commit()

    def __repr__(self):
        if len(self) < 20:
            return 'sqlist.SQList([%s])' % ', '.join(map(repr, self))
        else:
            return 'sqlist.SQList([%s]...)' % ', '.join(map(repr, self[:20]))

    def __len__(self):
        result = self.cursor.execute(
            '''SELECT COUNT(*) FROM `data`;'''
        ).fetchone()
        return result[0]

    def __getitem__(self, index):
        if type(index) == int:
            offset, stop, stride = slice(index, index + 1).indices(len(self))
        elif type(index) == slice:
            offset, stop, stride = index.indices(len(self))
        else:
            raise TypeError('Int or slice expected, %s found' % type(index))
        limit = stop - offset

        result = self.cursor.execute(
                '''SELECT `value`
                   FROM `data`
                   ORDER BY `key` ASC
                   LIMIT ? OFFSET ?;''',
                (limit, offset)
        )

        if type(index) == int:
            result = result.fetchone()
            if result is None:
                raise IndexError('%s is out of range' % index)
            else:
                return pickle.loads(result[0])
        else:
            return [pickle.loads(i[0]) for i in result]

    def __out_of_range_check(self, index, result):
        if not result.rowcount:
            self.sql.rollback()
            raise IndexError('%s is out of range' % index)
        else:
            self.sql.commit()

    def __setitem__(self, index, value):
        offset, stop, stride = slice(index, index + 1).indices(len(self))
        result = self.cursor.execute(
            '''UPDATE `data`
               SET `key` = ?, `value` = ?
               WHERE `_rowid_` = (
                  SELECT `_rowid_`
                  FROM `data`
                  ORDER BY `key` ASC
                  LIMIT 1 OFFSET ?
               );''',
            (self.key(value), pickle.dumps(value), offset)
        )
        self.__out_of_range_check(index, result)

    def __delitem__(self, index):
        offset, stop, stride = slice(index, index + 1).indices(len(self))
        result = self.cursor.execute(
            '''DELETE FROM `data`
               WHERE `_rowid_` = (
                  SELECT `_rowid_`
                  FROM `data`
                  ORDER BY `key` ASC
                  LIMIT 1 OFFSET ?
               );''',
            (offset, )
        )
        self.__out_of_range_check(index, result)

    def __iter__(self):
        for item in self.cursor.execute(
                '''SELECT `value` FROM `data` ORDER BY `key` ASC;'''):
            yield pickle.loads(item[0])
        raise StopIteration

    def __contains__(self, item):
        result = self.cursor.execute(
            '''SELECT `_rowid_`
               FROM `data`
               WHERE `value` = ?;''',
            (pickle.dumps(item),)
        )
        return bool(result.fetchone())

    def __eq__(self, other):
        if not hasattr(other, '__len__') and len(self) != len(other):
                return False

        this = self.cursor.execute(
                '''SELECT `value` FROM `data` ORDER BY `key` ASC;''')
        for a, b in zip((pickle.loads(i[0]) for i in this), other):
            if a != b:
                return False
        return True

    @classmethod
    def temp(cls, key=None, drop=False, suffix='', prefix='', dir=None):
        handle, path = tempfile.mkstemp(suffix, prefix, dir)
        close(handle)
        return cls(path=path, key=key, drop=drop)

    def extend(self, other):
        if isinstance(other, Iterable):
            self.cursor.executemany(
                '''INSERT INTO `data`
                   (`key`, `value`)
                   VALUES (?, ?);''',
                zip(map(self.key, other), map(pickle.dumps, other))
            )
            self.sql.commit()
        else:
            raise TypeError('unsupported operand type(s) for +: '
                            '\'sqlist.SQList\' and \'%s\'' % type(other))

    def append(self, value):
        self.cursor.execute(
            '''INSERT INTO `data`
               (`key`, `value`)
               VALUES (?, ?);''',
            (self.key(value), pickle.dumps(value))
        )
        self.sql.commit()

    def pop(self, index=-1):
        offset, stop, stride = slice(index, index + 1).indices(len(self))

        self.cursor.execute('BEGIN TRANSACTION;')
        result = self.cursor.execute(
            '''SELECT `_rowid_`, `value`
               FROM `data`
               ORDER BY `key` ASC
               LIMIT 1 OFFSET ?;''',
            (offset, )
        )
        if result.rowcount:
            rowid, value = result.fetchone()
            self.cursor.execute(
                '''DELETE FROM `data` WHERE `_rowid_` = ?;''', (rowid, )
            )
            self.sql.commit()
            return pickle.loads(value)
        else:
            self.sql.rollback()
            raise IndexError('{} is out of range'.format(index))

    def sort(self, key=None, reverse=False):
        def swap_required(a, b):
            """Check if items should be swapped considering reverse order"""
            try:
                if reverse:
                    return a < b
                else:
                    return a > b
            except TypeError as e:
                # Values are not comparable
                self.sql.rollback()
                raise TypeError(e)

        if key:
            if not callable(key):
                raise TypeError('{} object is not callable'.format(type(key)))
        else:
            def key(x):
                return x

        self.cursor.execute('BEGIN TRANSACTION;')
        if self.key:
            self.cursor.execute('''UPDATE `data` SET `key` = NULL;''')
            self.key = None
        position = 0
        while True:
            values = self.cursor.execute(
                '''SELECT `_rowid_`, `value` FROM `data` LIMIT 2 OFFSET ?''',
                (position, )
            ).fetchall()
            if len(values) < 2:
                break
            unpacked = [pickle.loads(i[1]) for i in values]
            q = ((values[1][1], values[1][0]), (values[0][1], values[0][0]))
            if swap_required(key(unpacked[0]), key(unpacked[1])):
                self.cursor.executemany(
                        '''UPDATE `data` SET `value` = ? WHERE `_rowid_` = ?''',
                        ((values[1][1], values[0][0]),
                         (values[0][1], values[1][0]))
                )
                if position:
                    position -= 1
            else:
                position += 1
        self.sql.commit()
